import contextlib
import functools
import operator


def product(iterable):
	return functools.reduce(operator.mul, iterable, 1)


def all_equal(iterable):
	try:
		value = first(iterable)
	except ValueError:
		return True
	for v in iterable:
		if v != value:
			return False
	return True


def one(iterable):
	"""Return the single element in iterable.

	Raise an error if there isn't exactly one element.
	"""
	item = None
	iterator = iter(iterable)
	try:
		item = next(iterator)
	except StopIteration:
		raise ValueError('Iterable is empty, must contain one item')
	try:
		next(iterator)
	except StopIteration:
		return item
	else:
		raise ValueError('object contains >1 items, must contain exactly one.')


def first(iterable):
	for item in iterable:
		return item
	raise ValueError('Iterable is empty')


def last(iterable):
	return first(reversed(iterable))


def nth(iterable, n, key=None):
	if n < 0:
		iterable = reversed(iterable)
		n = -n - 1
	for index, elem in enumerate(iterable):
		if index == n:
			return elem
	raise ValueError('Iterable is not long enough')


def count(iterable):
	with contextlib.suppress(TypeError):
		return len(iterable)
	i = 0
	for elem in iterable:
		i += 1
	return i


def match_any(res, string):
	"""Return if a string matches any of a list of regular expressions.

	Args:
		string (str): The string that you want to check.
		res: An Iterable of regular expressions to check.

	Returns:
		True if any of the regexes in res matches string, otherwise False.
	"""
	for regex in res:
		if regex.match(string):
			return True
	return False



"""This module provides a function `url_for_obj` to get
the correct URL for a given object. Classes are registered to endpoints to
handle objects of that type using a decorator. Parameters for the view are
taken from attributes of the object or mapped in the decorator.

Let's say you have a class:

class Employee(db.Model):
	__tablename__ = 'employees'

	id = Column(Integer, primary_key=True)

And a view for an individual employee:

@blueprint.route('/employee/<int:id>')
@url_for_obj.register(Employee)
def employee(id):
	employee = Employee.query.get_or_404(id)
	return render_template('employee.html', employee=employee)

You can now call `url_for_obj(some_employee)` instead of
`url_for('.employee', id=some_employee.id)`

By default, endpoint arguments will be taken from attributes
of the passed object. You can also specify a mapping of how to
generate the parameters from the object.

@blueprint.route('/employee/<dept>/<employee_name>')
@url_for_obj.register(models.Employee, {
	'dept': lambda employee: employee.department.name,
	'employee_name': lambda employee: employee.name,
	})
def employee(dept, employee_name):
	...

Add the url_for_obj function to the template context so you can use it in
templates.
Where your blueprint is defined:

@blueprint.context_processor
def context_processor():
	return {
		'url_for_obj': url_for_obj.url_for_obj,
		}

"""

# TODO make this an extension
# 	add url_for_obj to context in extension init_app
# 	store class_function_mapping in app context
# 	Don't assume it's part of a blueprint by adding '.' before the function name
# 	  Is there a way of checking? Should it be an option in register?

from inspect import signature

from flask import url_for

class_function_mapping = {}


def url_for_obj(obj, option=None):
	obj_type = type(obj)
	if option:
		key = (obj_type, option)
	else:
		key = obj_type
	try:
		render_func, get_funcs = class_function_mapping[key]
	except KeyError:
		raise ValueError('No view function registered for {}'.format(key))
	kwargs = {}
	for arg in signature(render_func).parameters:
		if arg in get_funcs:
			kwargs[arg] = get_funcs[arg](obj)
		else:
			kwargs[arg] = getattr(obj, arg)
	return url_for('.' + render_func.__name__, **kwargs)


def register(class_, option=None, get_funcs={}):
	"""A decorator to register a function as the way to display an object of class_
	"""
	if option:
		key = (class_, option)
	else:
		key = class_

	def decorator(func):
		class_function_mapping[key] = (func, get_funcs)
		return func
	return decorator



"""Utils for working with SQLAlchemy."""
import csv
from contextlib import suppress
from decimal import Decimal
import fractions
import logging
import os.path
from typing import Sequence
import uuid

from flask import abort
import flask_sqlalchemy
import pytz
from sqlalchemy import (
	Column,
	Integer,
	String,
	ForeignKey,
	DateTime,
	TypeDecorator,
	Numeric,
	ForeignKeyConstraint,
	)
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.dialects.postgresql.base import ischema_names
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.inspection import inspect
from sqlalchemy.orm.exc import MultipleResultsFound, NoResultFound
from sqlalchemy.orm.collections import (
	InstrumentedList,
	InstrumentedSet,
	InstrumentedDict,
	)
from sqlalchemy.types import Concatenable, UserDefinedType

logger = logging.getLogger(__name__)
db = flask_sqlalchemy.SQLAlchemy()


class CIText(Concatenable, UserDefinedType):
	"""Class to use Postgres' CIText column in SQLAlchemy.

	From: https://github.com/mahmoudimus/sqlalchemy-citext
	"""

	def get_col_spec(self):
		return 'CITEXT'

	def bind_processor(self, dialect):
		def process(value):
			return value
		return process

	def result_processor(self, dialect, coltype):
		def process(value):
			return value
		return process


# Register CIText to SQLAlchemy's Postgres reflection subsystem.
ischema_names['citext'] = CIText


def psycopg_uri(username, password, db_name, host='', port=None):
	"""Create a URL for psycopg2.

	http://docs.sqlalchemy.org/en/latest/dialects/postgresql.html#module-sqlalchemy.dialects.postgresql.psycopg2  # noqa
	"""
	if port:
		return f'postgresql+psycopg2://{username}:{password}@{host}:{port}/{db_name}'
	else:
		return f'postgresql+psycopg2://{username}:{password}@{host}/{db_name}'


def foreign_key_col(col, **kwargs):
	"""Create a column with a ForeignKey that cascades."""
	return Column(
		col.type,
		ForeignKey(col, ondelete='CASCADE', onupdate='CASCADE'),
		**kwargs
		)


def parent_key(column, col_type=Integer, nullable=False, index=True, **kwargs):
	"""Shorthand for a Column that is a foreign key to a parent."""
	return Column(
		col_type,
		ForeignKey(column, ondelete='CASCADE', onupdate='CASCADE'),
		nullable=nullable,
		index=index,
		**kwargs
		)


class Fraction(TypeDecorator):
	"""Type for storing and retrieving Fractions.

	Currently, this is backed by a Decimal, so some precision may be lost on
	conversion.
	"""

	impl = Numeric

	def process_bind_param(self, value, dialect):
		if value is None:
			return None
		assert isinstance(value, int) or isinstance(value, fractions.Fraction)
		return value.numerator / Decimal(value.denominator)

	def process_result_value(self, value, dialect):
		if value is None:
			return None
		assert isinstance(value, Decimal)
		return fractions.Fraction(value).limit_denominator()


class UTCDateTime(TypeDecorator):
	"""Type for storing and retrieving DateTimes as UTC.

	Naive datetimes are rejected.
	The returned value is always a non-naive UTC datetime.
	"""

	impl = DateTime

	def process_bind_param(self, value, dialect):
		"""Fix inputs."""
		if value is None:
			return None
		if value.tzinfo is None:
			raise ValueError
		return value.astimezone(pytz.utc)

	def process_result_value(self, value, dialect):
		"""Normalize output."""
		if value is None:
			return None
		return value.replace(tzinfo=pytz.utc)


class BaseMixin():
	# Build table args using inheritance

	_table_args = ()
	_table_kwargs = {}

	@declared_attr
	def __table_args__(cls):
		args = []
		kwargs = {}
		for par_cls in reversed(cls.mro()):
			with suppress(AttributeError):
				args.extend(par_cls._table_args)
			with suppress(AttributeError):
				kwargs.update(par_cls._table_kwargs)
		return tuple(args) + (kwargs,)

	@classmethod
	def find_one(cls, **kwargs):
		"""Query this table for a single row matching kwargs filters."""
		return cls.query.filter_by(**kwargs).one()

	@classmethod
	def find_one_or_404(cls, **kwargs):
		"""Query this table for a single row, flask.abort(404) if not found."""
		try:
			cls.find_one(**kwargs)
		except (NoResultFound, MultipleResultsFound):
			abort(404)

	@classmethod
	def create(cls, **kwargs):
		obj = cls(**kwargs)
		cls.query.session.add(obj)
		return obj

	@classmethod
	def find_create(cls, create_args=None, **kwargs):
		"""Find or create an instance of this model.

		Optionally provide arguments used only for creating the object, not
		querying.
		"""
		try:
			return cls.find_one(**kwargs)
		except NoResultFound:
			create_args = dict(create_args or {})
			create_args.update(kwargs)
			return cls.create(**create_args)

	@classmethod
	def exists(cls, **kwargs):
		try:
			cls.find_one(**kwargs)
		except NoResultFound:
			return False
		else:
			return True

	@classmethod
	def cast(cls, obj, allow_none=True):
		if obj is None:
			if allow_none:
				return None
			else:
				raise ValueError(f'Cannot cast None to {cls}')
		elif isinstance(obj, cls):
			return obj
		else:
			raise TypeError(f'Cannnot cast an object of {type(obj)} to {cls}')

	@classmethod
	def _repr_class_template(cls):
		col_names = [col.name for col in cls.__table__.columns]
		item_format = '{col}={{self.{col}!r}}'
		fields = ', '.join(item_format.format(col=col) for col in col_names)
		return '{class_name}({fields})'.format(
			class_name=cls.__name__,
			fields=fields,
			)

	def __repr__(self):
		class_template = self._repr_class_template()
		return class_template.format(obj=self, self=self)

	@classmethod
	def query_default_order(cls):
		"""Return a class query with a default (stable) order."""
		primary_key_cols = inspect(cls).primary_key
		cls.query.order_by(*primary_key_cols)

	@classmethod
	def _get_pkey_col(cls):
		primary_key_cols = inspect(cls).primary_key
		if len(primary_key_cols) != 1:
			msg = 'Class %s does not have exactly one primary key column' % cls
			raise NotImplementedError(msg)
		return primary_key_cols[0]

	@classmethod
	def fkey_constraint(
		cls, cols: Sequence,
		ondelete='CASCADE',
		onupdate='CASCADE',
		) -> ForeignKeyConstraint:
		"""Return a ForeignKeyConstraint for the primary keys of this model."""
		return ForeignKeyConstraint(
			cols,
			inspect(cls).primary_key,
			ondelete=ondelete,
			onupdate=onupdate,
			)

	@classmethod
	def pkey(cls, **kwargs):
		"""Return a Column definition for the primary key of this model."""
		pkey_col = cls._get_pkey_col()
		return foreign_key_col(pkey_col, **kwargs)

	def to_dict(self, sub=None):
		"""Create a dict of this obj's attributes and optionally related objects.

		Specify which relationships to include as keys in `sub`. The value of
		each key is the argument to pass to that relation's `to_dict`.

		For example:

		class Character(BaseModel):
			first_name = Column(String)
			last_name = Column(String)
			parents = relationship(Character)
			siblings = relationship(Character)

		`bart.to_dict({'parents': {'siblings': None}}) == {
			'first_name': 'Bart',
			'last_name': 'Simpson',
			'parents': [
				{
					'first_name': 'Homer',
					'last_name': 'Simpson',
					'siblings': [],
					},
				{
					'first_name': 'Marge',
					'last_name': 'Simpson',
					'siblings': [
						{
							'first_name': 'Patty',
							'last_name': 'Bouvier',
							},
						{
							'first_name': 'Selma',
							'last_name': 'Bouvier',
							},
						],
					},
				]
			}`
		"""
		instance_state = inspect(self)
		columns = instance_state.mapper.column_attrs
		out = {col.key: getattr(self, col.key) for col in columns}
		sub = sub or {}
		for relationship_name, args in sub.items():
			relationship = getattr(self, relationship_name)
			try:
				out[relationship_name] = relationship.to_dict(args)
			except AttributeError:
				if isinstance(relationship, (InstrumentedList, InstrumentedSet)):
					out[relationship_name] = []
					for relation in relationship:
						out[relationship_name].append(relation.to_dict(args))
				elif isinstance(relationship, InstrumentedDict):
					out[relationship_name] = {}
					for k, relation in relationship.items():
						out[relationship_name][k] = relation.to_dict(args)
				else:
					msg = "Don't know how to handle relationship of type {rel_type}".format(
						rel_type=type(relationship),
						)
					raise NotImplementedError(msg)
		return out

	@classmethod
	def _get_path(cls, path, directory):
		if path is None:
			path = '%s.csv' % cls.__tablename__
			if directory:
				path = os.path.join(directory, path)
		else:
			if directory:
				raise ValueError('Must not pass path and directory')
		return path

	@classmethod
	def merge_csv(cls, path=None, directory=None, io_wrapper=None):
		"""Load data from a csv file and merge into db.

		This is slower than import because orm objects are created.
		"""
		path = cls._get_path(path, directory)
		logger.info('Importing %s to %s', path, cls)
		with open(path, 'r', newline='') as infile:
			cls.merge_from_file(infile, io_wrapper=io_wrapper)

	@classmethod
	def import_csv(cls, path=None, directory=None, io_wrapper=None):
		"""Import data from a csv file."""
		path = cls._get_path(path, directory)
		logger.info('Importing %s to %s', path, cls)
		with open(path, 'r', newline='') as infile:
			cls.import_from_file(infile, io_wrapper=io_wrapper)

	@classmethod
	def export_csv(cls, path=None, directory=None, io_wrapper=None):
		"""Export data to csv file."""
		path = cls._get_path(path, directory)
		logger.info('Exporting %s to %s', cls, path)
		try:
			with open(path, 'r', newline='') as infile:
				reader = csv.DictReader(infile)
				fieldnames = reader.fieldnames
		except IOError:
			fieldnames = None
		with open(path, 'w', newline='') as outfile:
			cls.export_to_file(outfile, fieldnames=fieldnames, io_wrapper=io_wrapper)

	@classmethod
	def import_from_file(cls, infile, io_wrapper=None):
		logger.info('Importing to %s', cls)
		if io_wrapper:
			infile = io_wrapper(infile)
		reader = csv.DictReader(infile)
		# TODO try insert many (not on session)
		# http://stackoverflow.com/questions/25694234/bulk-update-in-sqlalchemy-core-using-where
		# db.session.bulk_insert_mappings(cls, reader)
		rows = list(reader)
		db.engine.execute(cls.__table__.insert(), rows)

	@classmethod
	def merge_from_file(cls, infile, io_wrapper=None):
		logger.info('Merging to %s', cls)
		if io_wrapper:
			infile = io_wrapper(infile)
		reader = csv.DictReader(infile)
		for row in reader:
			obj = cls(**row)
			cls.query.session.merge(obj)

	@classmethod
	def export_to_file(cls, outfile, fieldnames=None, io_wrapper=None):
		logger.info('Exporting %s', cls)
		if io_wrapper:
			outfile = io_wrapper(outfile)
		fieldnames = fieldnames or []
		for col in cls.__table__.columns:
			if col.name not in fieldnames:
				fieldnames.append(col.name)
		writer = csv.DictWriter(outfile, fieldnames, extrasaction='ignore')
		writer.writeheader()
		try:
			query = cls.query_default_order()
		except NotImplementedError:
			query = cls.query
		for obj in query:
			writer.writerow(obj.__dict__)


class IntegerPKey():
	"""Mixin for models with an integer 'id' as the primary key."""

	id = Column(Integer, primary_key=True)

	@classmethod
	def cast(cls, obj, **kwargs):
		if isinstance(obj, int):
			value = cls.query.get(obj)
			if not value:
				raise ValueError(f"No {cls} could be found from int {obj}")
			return value
		else:
			return super().cast(obj, **kwargs)

	@classmethod
	def _get_pkey_col(cls):
		return cls.id

	def __hash__(self):
		return self.id

	def __eq__(self, other):
		return self.__class__ == other.__class__ and self.id == other.id

	@classmethod
	def query_default_order(cls):
		return cls.query.order_by(cls.id)


def parse_uuid(obj) -> uuid.UUID:
	"""Parse a UUID from obj.

	Raises:
		ValueError: If obj cannot be parsed into a UUID
		TypeError: If the type of obj isn't known how to parse
	"""
	# TODO this would be more pythonic with try except instead of checking types
	if isinstance(obj, uuid.UUID):
		return obj
	if isinstance(obj, str):
		return uuid.UUID(obj)
	if isinstance(obj, (bytes, bytearray)):
		return uuid.UUID(bytes=obj)
	if isinstance(obj, tuple):
		return uuid.UUID(fields=obj)
	if isinstance(obj, int):
		return uuid.UUID(int=obj)
	raise TypeError


class UUID4Pkey():
	"""Mixin for models with a UUID4 'id' as the primary key."""

	id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)

	@classmethod
	def cast(cls, obj, **kwargs):
		try:
			obj_as_uuid = parse_uuid(obj)
		except (ValueError, TypeError):
			return super().cast(obj, **kwargs)
		else:
			value = cls.query.get(obj_as_uuid)
			if not value:
				raise ValueError(f"No {cls} could be found from uuid {obj}")
			return value

	@classmethod
	def _get_pkey_col(cls):
		return cls.id

	def __hash__(self):
		return hash(self.__class__, self.id)

	def __eq__(self, other):
		return self.__class__ == other.__class__ and self.id == other.id

	@classmethod
	def query_default_order(cls):
		return cls.query.order_by(cls.id)


class AbbrPKey():

	abbr = Column(String, primary_key=True)

	@classmethod
	def cast(cls, obj, **kwargs):
		if isinstance(obj, str):
			value = cls.query.get(obj)
			if not value:
				raise ValueError(f"No {cls} could be found from str {obj}")
			return value
		else:
			return super().cast(obj, **kwargs)

	@classmethod
	def _get_pkey_col(cls):
		return cls.abbr

	def __hash__(self):
		return hash(self.abbr)

	def __eq__(self, other):
		return self.__class__ == other.__class__ and self.abbr == other.abbr

	@classmethod
	def query_default_order(cls):
		return cls.query.order_by(cls.abbr)


class no_autoflush(object):
	"""A context manager to suppress autoflush."""
	def __init__(self, session):
		self.session = session
		self.autoflush = session.autoflush

	def __enter__(self):
		self.session.autoflush = False

	def __exit__(self, type, value, traceback):
		self.session.autoflush = self.autoflush



"""Mixins to implement rich comparison based on a key."""
import abc


class ComparableMixin():
	"""
	A class inheriting from ComparableMixin must implement only two functions to
	get all comparison methods.

	`_cmp_key` returns a key for comparing the objects.
	`_is_valid_operand` takes another object and determines if this class knows
		how to compare them.
	"""
	# TODO this should have a metaclass of abc.ABCMeta but I don't feel like
	# dealing with the metaclass conflict when using this with other Mixins that
	# have metaclasses (eg. SQLAlchemy models)

	@classmethod
	def _is_valid_operand(cls, other):
		raise NotImplementedError()

	@abc.abstractmethod
	def _cmp_key(self):
		raise NotImplementedError()

	def __eq__(self, other):
		if not self._is_valid_operand(other):
			return False
		return self._cmp_key() == other._cmp_key()

	def __ne__(self, other):
		if not self._is_valid_operand(other):
			return True
		return self._cmp_key() != other._cmp_key()

	def __gt__(self, other):
		if not self._is_valid_operand(other):
			return NotImplemented
		return self._cmp_key() > other._cmp_key()

	def __ge__(self, other):
		if not self._is_valid_operand(other):
			return NotImplemented
		return self._cmp_key() >= other._cmp_key()

	def __lt__(self, other):
		if not self._is_valid_operand(other):
			return NotImplemented
		return self._cmp_key() < other._cmp_key()

	def __le__(self, other):
		if not self._is_valid_operand(other):
			return NotImplemented
		return self._cmp_key() <= other._cmp_key()

	def __hash__(self):
		return hash(self._cmp_key())


class ComparableIsInstanceMixin(ComparableMixin):
	"""This class can compare to instances of the class."""

	@classmethod
	def _is_valid_operand(cls, other):
		return isinstance(other, cls)


class ComparableSameClassMixin(ComparableMixin):
	"""This class can compare to the same exact class."""

	@classmethod
	def _is_valid_operand(cls, other):
		return other.__class__ == cls



import collections
import csv

import openpyxl


def csv_to_dict(filepath, key_cols):
	"""
	Parse a csv file.  Return a dict with key_cols as the keys.  For example
	if the key_cols were ('A', 'B', 'C') the output would be:
	output[(A, B, C)] = fields
	where A, B, and C are the values of the respective columns on a given line and
	fields is a dict mapping the other column names to the values from that line.
	"""
	out = {}
	with open(filepath) as infile:
		reader = csv.DictReader(infile)
		for line in reader:
			values = dict(line)
			key = tuple([values.pop(key_col) for key_col in key_cols])
			out[key] = values
	return out


def sort_dict(d, key=None):
	"""Return an OrderedDict of d sorted."""
	out = collections.OrderedDict()
	for k in sorted(d, key=key):
		out[k] = d[k]
	return out


def group_dicts_by(dict_list, key):
	agged_data = collections.defaultdict(list)
	for d in dict_list:
		agged_data[d[key]].append(d)
	return agged_data


def xlsx_dict_reader(filepath):
	wb = openpyxl.load_workbook(filepath)
	ws = wb.active
	header_row = None
	for row in ws.rows:
		row_values = [cell.value for cell in row]
		if header_row is None:
			header_row = row_values
		else:
			yield dict(zip(header_row, row_values))



"""Render WTForms using Bootstrap 3.

Inspired by https://github.com/carlnewton/bootstrap-wtf

TODO:
* Don't depend on glyphicons. Maybe take an option for glyphicons, fontawesome,
	or none? We just need it for the close icon. We could fall back on an X.
	Icons depend on FontAwesome, should we just assume that?
* Help text aria-describedby
	https://getbootstrap.com/docs/3.3/css/#forms-help-text
* Disabled & Readonly
	https://getbootstrap.com/docs/3.3/css/#forms-control-disabled
	https://getbootstrap.com/docs/3.3/css/#forms-control-readonly
* Check types using inheritance instead of checking type names
* Rename TEXT_FIELD_TYPES to reflect that it is checking
	textual <input>, <textarea>, and <select> not just text fields
* Add support for static control "fields"
	https://getbootstrap.com/docs/3.3/css/#forms-controls-static
"""
from typing import Dict, Sequence, Union, Mapping, Any, Optional

from flask import Markup
from wtforms import Form, Field, FieldList, HiddenField


TEXT_FIELD_TYPES = frozenset((
	'DateField',
	'DateTimeField',
	'DecimalField',
	'FloatField',
	'IntegerField',
	'PasswordField',
	'SelectField',
	'SelectMultipleField',
	'StringField',
	'TextAreaField',
	'TextField',
	'SearchField',
	'TelField',
	'URLField',
	'EmailField',
	'TimeField',
	'DateTimeLocalField',
	'IntegerRangeField',
	'DecimalRangeField',
	'QuerySelectField',
	'QuerySelectMultipleField',
	))
OTHER_FIELD_TYPES = frozenset((
	'BooleanField',
	'FileField',
	'RadioField',
	'SubmitField',
	))
KNOWN_FIELD_TYPES = TEXT_FIELD_TYPES & OTHER_FIELD_TYPES
FIELD_TYPE_ICONS = {
	'EmailField': '@',
	'DateField': '<span class="fa fa-calendar"></span>',
	'DateTimeField': '<span class="fa fa-calendar"></span>',
	'TelField': '<span class="fa fa-phone"></span>',
	'TimeField': '<span class="fa fa-clock-o"></span>',
	'PasswordField': '<span class="fa fa-key"></span>',
	'URLField': '<span class="fa fa-link"></span>',
	}
ERROR_ICON = (
	'<span class="glyphicon glyphicon-remove form-control-feedback" '
	'aria-hidden="true"></span><span class="sr-only">(error)</span>'
	)


def _space_join_cond(names_conditions: Mapping[Optional[str], Any]) -> str:
	"""Return a space joined string based on mapped conditions."""
	return ' '.join(k for k, v in names_conditions.items() if (k and v))


def _col(text: str, width: int, size: str = 'sm', offset: int = None) -> str:
	"""Wrap text in a column with width."""
	classes = _space_join_cond({
		f'col-{size}-{width}': True,
		f'col-{size}-offset-{offset}': offset,
		})
	return f'<div class="{classes}">{text}</div>'


def bootstrap_field(
		field: Field,
		form_group: bool = True,
		placeholder: Union[bool, str] = True,
		label: bool = True,
		errors: bool = True,
		horizontal: bool = False,
		inline: bool = False,
		btn_color: str = 'default',
		icon: Union[bool, str] = False,
		**kwargs
		) -> Markup:
	"""Render a WTForms Field with Bootstrap markup.

	Args:
		field: The field to render.
		form_group: Whether or not to make it a form-group.
		placeholder: The string to display as a placeholder, True to display the
			default placeholder and something falsey to display nothing.
		label: Whether or not to display the field label.
		errors: Whether or not to display errors.
		horizontal: Whether to display the field horizontally.
		icon:
		inline: Whether to display the field inline.
		btn_color: The color name to pass as the button color.
	"""
	# TODO respect inline
	if isinstance(field, HiddenField):
		# Handles CSRFTokenFields
		return Markup(str(field))

	is_text_field = field.type in TEXT_FIELD_TYPES
	field_classes = _space_join_cond({
		'form-control': is_text_field,
		'btn btn-' + btn_color: field.type == 'SubmitField',
		})
	if placeholder is True and field.type in TEXT_FIELD_TYPES:
		placeholder = 'Enter %s' % field.label.text
	html_str = field(
		class_=field_classes,
		placeholder=placeholder,
		**kwargs
		)
	html_str = _insert_icon(html_str, field, icon, is_text_field)

	horizontal_field_offset = 2
	label_html = ''
	if horizontal and field.errors:
		html_str += ERROR_ICON
	if field.type == 'BooleanField':
		html_str = (
			f'<div class="checkbox"><label>{html_str} {field.label.text}'
			'</label></div>'
			)
	elif field.type != 'SubmitField' and label:
		label_html = field.label(class_=_space_join_cond({
			'control-label': True,
			'col-sm-2': horizontal
			}))
		horizontal_field_offset = 0
	if horizontal:
		html_str = label_html + _col(html_str, 10, offset=horizontal_field_offset)
	else:
		html_str = label_html + html_str
		if form_group and field.errors:
			html_str += ERROR_ICON

	if form_group:
		form_group_classes = _space_join_cond({
			'form-group': True,
			'has-error has-feedback': field.errors,
			'required': field.flags.required,
			})
		html_str = f'<div class="{form_group_classes}">{html_str}</div>'

	html_str += _field_description(field, horizontal)
	html_str += _field_errors(field, errors, horizontal)

	return Markup(html_str)


def _insert_icon(
		html_str: str,
		field: Field,
		icon: Union[str, bool, None],
		is_text_field: bool,
) -> str:
	"""Insert an icon into a bootstrap input field."""
	if icon and is_text_field:
		if icon is True:
			icon = FIELD_TYPE_ICONS.get(field.type)
		if icon:
			html_str = (
				f'<div class="input-group"><span class="input-group-addon">'
				f'{icon}</span>{html_str}</div>'
				)
	return html_str


def _field_description(field, horizontal):
	description_html = ''
	if field.description:
		description_html = f'<p class ="help-block">{field.description}</p>'
		if horizontal:
			description_html = _col(description_html, 10, offset=2)
	return description_html


def _field_errors(field, errors, horizontal):
	error_html = ''
	if field.errors and errors:
		if len(field.errors) == 1:
			error_html = f'<p class="text-danger">{field.errors[0]}</p>'
		else:
			error_html = ''.join(
				f'<li class="text-danger">{error}</li>'
				for error in field.errors
				)
			error_html = f'<ul>{error_html}</ul>'
		if horizontal:
			error_html = _col(error_html, 10, offset=2)
	return error_html


def _gen_fields(obj: Union[Form, Sequence[Field], FieldList]):
	"""Recursively generate all of the fields in obj."""
	for field in obj:
		if field.type == 'FieldList':
			yield from _gen_fields(field)
		else:
			yield field


def bootstrap_form(
		form: Form,
		action: str = None,
		*,
		method: str = 'POST',
		form_groups: bool = True,
		placeholders: bool = True,
		labels: bool = True,
		errors: bool = True,
		horizontal: bool = False,
		inline: bool = False,
		icons: bool = True,
		id_: str = None,
		class_: str = None,
		field_args: Dict[str, Dict] = None,
) -> Markup:
	"""Render a WTForms Form with Bootstrap markup.

	Args:
		form: The form to render.
		action: The action of the form.
		method: The method of the form.
		form_groups: Whether to render form groups.
		placeholders: Whether to display placeholders.
		labels: Whether to display labels.
		errors: Whether to display errors.
		horizontal: Render horizontal Bootstrap form.
		inline: Render inline Bootstrap form.
		icons: Whether to display icons in the input.
		id_: An HTML id for the form.
		class_: Classes to add to the form.
		field_args: A mapping of field names to additional arguments to pass
			to each field.
	"""
	if horizontal and inline:
		raise ValueError
	field_args = field_args or {}

	file_upload = any(field.type == 'FileField' for field in _gen_fields(form))
	form_classes = _space_join_cond({
		'form-horizontal': horizontal,
		'form-inline': inline,
		class_: class_,
		})
	form_attrs = _space_join_cond({
		'id="%s"' % id_: id_,
		'method="%s"' % method: method,
		'action="%s"' % action: action,
		'enctype=multipart/form-data': file_upload,
		'class="%s"' % form_classes: form_classes,
		})

	field_values = []
	for field in _gen_fields(form):
		extra_args = field_args.get(field.name, {})
		field_values.append(bootstrap_field(
			field,
			form_group=form_groups,
			placeholder=placeholders,
			label=labels,
			errors=errors,
			horizontal=horizontal,
			inline=inline,
			icon=icons,
			**extra_args
			))
	fields_html = ''.join(field_values)
	html = f'<form {form_attrs}>{fields_html}</form>'
	return Markup(html)



"""Tools to build a URL using the current URL as defaults."""
from collections import Iterable

from flask import request, url_for
from werkzeug.datastructures import MultiDict


def url_update_args(**kwargs):
	"""Take the current URL and update only some parameters.

	Keeping all unspecified params the same and only update the passed parms.
	"""
	return url_update_endpoint_args(request.endpoint, **kwargs)


def url_update_endpoint_args(endpoint, **kwargs):
	"""Like url_for by using current URL for unspecified parameters.

	Return the URL for passed endpoint using args from current request and kwargs.
	"""
	# request.args contains parameters from the query string
	# request.view_args contains parameters that matched the view signature
	args = MultiDict(request.args)
	args.update(request.view_args)
	for arg, value in kwargs.items():
		if isinstance(value, Iterable) and not isinstance(value, str):
			args.setlist(arg, value)
		else:
			args[arg] = value
	# Now set any individual args to the object instead of a list of len 1
	args = args.to_dict(flat=False)
	for key in set(args.keys()):
		if len(args[key]) == 1:
			args[key] = args[key][0]
	return url_for(endpoint, **args)


def url_update(*args, **kwargs):
	if args:
		if len(args) > 1:
			raise ValueError('Can only pass one positional argument')
		return url_update_endpoint_args(args[0], **kwargs)
	else:
		return url_update_endpoint_args(request.endpoint, **kwargs)



from datetime import timedelta
import string


DESIGNATORS_DAYS = {
	'Y': timedelta(days=365),
	'M': timedelta(days=30),
	'W': timedelta(days=7),
	'D': timedelta(days=1),
	}
DESIGNATORS_TIME = {
	'H': timedelta(hours=1),
	'M': timedelta(minutes=1),
	'S': timedelta(seconds=1),
	}
AMOUNT_CHARS = frozenset(string.digits + '.,')
DAYS_DESIGNATORS = 'YMWD'
TIME_DESIGNATORS = 'HMS'


def parse_duration_iso(s: str) -> timedelta:
	"""Parse an ISO 8601 duration string to a timedelta.

	This is inexact because years and months aren't specific amounts of time.

	PnYnMnDTnHnMnS where the 'n's are integer amounts.

	https://www.wikiwand.com/en/ISO_8601#/Durations

	Raises:
		ValueError: If the string is malformed.
	"""
	if not s.startswith('P'):
		raise ValueError('Duration string must start with a P')
	if len(s) < 3:
		raise ValueError('Duration string must be at least 3 characters long')
	if 'T' in s:
		days_string, time_string = s[1:].split('T')
	else:
		days_string, time_string = s[1:], ''
	running_total = _parse_duration_days(days_string, bool(time_string))
	running_total += _parse_duration_time(time_string)
	return running_total


def _parse_duration_days(days_string, any_time_string):
	running_total = timedelta(0)
	amount_string = ''
	last_designator = None
	for char in days_string:
		if char in AMOUNT_CHARS:
			amount_string += char
		else:
			if not amount_string:
				raise ValueError('Empty amount')
			# Is this the final value?
			if not any_time_string and char == days_string[-1]:
				amount = float(amount_string.replace(',', '.'))
			else:
				amount = int(amount_string)
			try:
				designator_value = DESIGNATORS_DAYS[char]
			except KeyError:
				raise ValueError('Unknown designator "%s"' % char)
			if last_designator:
				designator_index = DAYS_DESIGNATORS.index(char)
				last_designator_index = DAYS_DESIGNATORS.index(last_designator)
				if designator_index <= last_designator_index:
					raise ValueError(
						'Designators appear out of order, like minutes before '
						'hours'
						)
			running_total = running_total + (designator_value * amount)
			amount_string = ''
			last_designator = char
	return running_total


def _parse_duration_time(time_string):
	running_total = timedelta(0)
	amount_string = ''
	last_designator = None
	for char in time_string:
		if char in AMOUNT_CHARS:
			amount_string += char
		else:
			if not amount_string:
				raise ValueError('Empty amount')
			# Is this the final value?
			if char == time_string[-1]:
				amount = float(amount_string.replace(',', '.'))
			else:
				amount = int(amount_string)
			try:
				designator_value = DESIGNATORS_TIME[char]
			except KeyError:
				raise ValueError('Unknown designator "%s"' % char)
			if last_designator:
				designator_index = TIME_DESIGNATORS.index(char)
				last_designator_index = TIME_DESIGNATORS.index(last_designator)
				if designator_index <= last_designator_index:
					raise ValueError(
						'Designators appear out of order, like minutes before '
						'hours'
						)
			running_total = running_total + (designator_value * amount)
			amount_string = ''
			last_designator = char
	return running_total



from datetime import date, datetime
from typing import List, Iterable
from uuid import UUID

from werkzeug.routing import BaseConverter, ValidationError


class DateConverter(BaseConverter):
	"""URL converter for dates."""

	def to_python(self, value: str) -> date:
		"""Turn a string into a date."""
		try:
			return datetime.strptime(value, '%Y-%m-%d').date()
		except ValueError:
			raise ValidationError()

	def to_url(self, d: date) -> str:
		"""Format a date for a URL."""
		return d.isoformat()


def ListConverter(char: str) -> type(BaseConverter):
	"""Factory to create a Converter that turns delimited strings into lists."""

	class _ListConverter(BaseConverter):
		f"""URL converter for Iterable <-> '{char}' delimited string."""

		def to_python(self, value: str) -> List[str]:
			"""Create a tuple from char delimited values."""
			return value.split(char)

		def to_url(self, value: Iterable[str]) -> str:
			"""Serialize Iterable to char delimited values."""
			return char.join(value)

	return _ListConverter


class UUIDConverter(BaseConverter):
	"""URL converter for UUIDs."""

	def to_python(self, value: str) -> UUID:
		"""Create the UUID from string."""
		return UUID(value)

	def to_url(self, value: UUID) -> str:
		"""Turn a UUID into a string."""
		return str(value)



"""Parse different types of strings."""

import decimal


def money(money_string: str) -> decimal.Decimal:
	"""Parse a string representing some amount of money into a Decimal.

	Args:
		money_string (str): The string to try to parse

	Returns:
		A Decimal that is the amount of money.
	"""
	if not money_string:
		return decimal.Decimal(0)
	if money_string.startswith('$'):
		money_string = money_string[1:]
	money_string = money_string.replace(',', '')
	return decimal.Decimal(money_string)


def accounting(s: str) -> decimal.Decimal:
	"""Parse a string using accounting format (parenthesis for neg amounts)."""
	s = s.strip()
	if not s or s == '-':
		return decimal.Decimal(0)
	if s[0] == '(' and s[-1] == ')':
		s = '-' + s[1:-1]
	return decimal.Decimal(s)


def percent_to_float(s: str) -> float:
	"""Parse a percentage string into a float."""
	s = s.strip()
	if s.endswith('%'):
		s = s[:-1]
	return float(s) / 100.0


def percent_to_decimal(s: str, blank_return_value=None) -> decimal.Decimal:
	s = s.strip()
	if s.endswith('%'):
		s = s[:-1]
	if s == '':
		return blank_return_value
	return decimal.Decimal(s) / 100



"""click interface for postgresql commands.

"""
import logging
import os
import subprocess

import click
from flask import current_app
from flask.cli import AppGroup

DEFAULT_DB_BACKUP_PATH = 'db.pg_dump'

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

cli = AppGroup(help=__doc__, name='pg')


@cli.command()
@click.option('-l', '--location', default=DEFAULT_DB_BACKUP_PATH)
@click.option('--format', default='c')
def dump(location, format):
	"""Run pg_dump."""
	os.environ['PGPASSWORD'] = current_app.config['PG_PASSWORD']
	pg_dump = current_app.config.get('PG_BIN_DIR', '') + 'pg_dump'
	subprocess.call((
		pg_dump,
		'--host={}'.format(current_app.config['PG_HOST']),
		'--username={}'.format(current_app.config['PG_USERNAME']),
		'--format=%s' % format,
		current_app.config['PG_DB_NAME'],
		'--file=%s' % location,
		))


@cli.command()
@click.option('-l', '--location', default=DEFAULT_DB_BACKUP_PATH)
def restore(location):
	"""Restore pg_dump."""
	os.environ['PGPASSWORD'] = current_app.config['PG_PASSWORD']
	pg_restore = current_app.config.get('PG_BIN_DIR', '') + 'pg_restore'
	subprocess.call((
		pg_restore,
		'--host={}'.format(current_app.config['PG_HOST']),
		'--username={}'.format(current_app.config['PG_USERNAME']),
		'--dbname={}'.format(current_app.config['PG_DB_NAME']),
		'--clean',
		location,
		))


@cli.command()
@click.option('-c', '--command')
@click.option('-f', '--file')
def psql(command=None, file=None):
	"""Call psql."""
	os.environ['PGPASSWORD'] = current_app.config['PG_PASSWORD']
	psql = current_app.config.get('PG_BIN_DIR', '') + 'psql'
	args = [
		psql,
		'--host={}'.format(current_app.config['PG_HOST']),
		'--username={}'.format(current_app.config['PG_USERNAME']),
		'--dbname={}'.format(current_app.config['PG_DB_NAME']),
		]
	if command:
		args.append('--command=%s' % command)
	if file:
		args.append('--file=%s' % file)
	subprocess.call(args)
